#!/usr/bin/python
# coding: utf8

import pathlib
import subprocess
import tempfile

QRC_TEMPLATE = """
<!DOCTYPE RCC><RCC version="1.0">
<qresource>
{files}
</qresource>
</RCC>
"""

CURDIR = pathlib.Path(__file__).parent
COMPILER = "pyside6-rcc"
RESOURCES = CURDIR / "resources"
DESTINATION = CURDIR / "tankobon" / "gui" / "resources.py"


def create_qrc():
    return QRC_TEMPLATE.format(
        files="\n".join(
            f'<file alias="{file.name}">{str(file)}</file>'
            for file in RESOURCES.glob("*")
            if not file.name.startswith(".")
        )
    ).strip()


def compile_qrc(path_to_qrc, dest):
    subprocess.run([COMPILER, path_to_qrc, "-o", dest], check=True)


if __name__ == "__main__":
    with tempfile.NamedTemporaryFile(suffix=".rc") as f:
        f.write(create_qrc().encode("utf8"))
        f.seek(0)

        compile_qrc(f.name, str(DESTINATION))



# coding: utf8

import os
import platform
import subprocess

from PySide6.QtCore import QResource
from PySide6.QtGui import QColor

SYSTEM = platform.system()


def resource(path: str) -> bytes:
    return QResource(path).data().tobytes()  # type: ignore


def is_dark(color: QColor) -> bool:
    hex_color = color.name().lstrip("#")
    colors = [int(hex_color[i : i + 2], 16) for i in (0, 2, 4)]
    return [color < 0x7F for color in colors].count(True) >= 2


def xopen(path: str):
    """Open a path using the correct application.

    Args:
        path: The path to open.
    """

    if SYSTEM == "Windows":
        os.startfile(path)  # type: ignore
    elif SYSTEM == "Darwin":
        subprocess.run(["open", path])
    else:
        subprocess.run(["xdg-open", path])



# coding: utf8

import json
from typing import Dict

from .. import core, models

# from ..exceptions import MangaError

METADATA_MAP = {
    "title": "title",
    "alt_titles": "alt_titles",
    "authors": "authors",
    "genres": "genres",
    "description": "desc",
}


class Parser(core.Parser):

    domain = r"catmanga.org/series/(\w+)"

    # keep state
    # FIXME: cache may grow too large when requesting many manga
    _cache: Dict[str, dict] = {}

    def _get_data(self, url):

        if url not in self._cache:
            soup = self.soup(url)
            # catmanga happens to be a next.js app, so we can use this JSON data.
            # https://github.com/vercel/next.js/discussions/15117
            data = soup.find("script", id="__NEXT_DATA__").string

            self._cache[url] = json.loads(data)["props"]["pageProps"]["series"]

        return self._cache[url]

    def metadata(self, url):
        data = self._get_data(url)

        return models.Metadata(
            url=url,
            cover=data["cover_art"]["source"],
            **{METADATA_MAP[k]: v for k, v in data.items() if k in METADATA_MAP},
        )

    def add_chapters(self, manga):
        data = self._get_data(manga.meta.url)

        for cdata in data["chapters"]:
            cid = str(cdata["number"])

            manga.add(
                models.Chapter(
                    id=cid,
                    url=f"{manga.meta.url}/{cid}",
                    title=cdata.get("title") or "",
                )
            )

    def add_pages(self, chapter):
        soup = self.soup(chapter.url)
        cdata = json.loads(soup.find("script", id="__NEXT_DATA__").string)

        chapter.pages = cdata["props"]["pageProps"]["pages"]



# coding: utf8
"""Utilities for tankobon."""

import collections
import gzip
import json
import logging
import os
import pathlib
import re
from typing import Optional, Union
from urllib.parse import urlparse

import bs4  # type: ignore
import fake_useragent as ua  # type: ignore
import filetype  # type: ignore
import requests

Number = Union[int, float]

_log = logging.getLogger("tankobon")

BS4_PARSER = "html5lib"  # if you want, change to lxml for faster parsing
USER_AGENT = ua.UserAgent()

RE_DOMAIN = re.compile(r"^(?:www\.)?(.*)(:(\d+))?$")

# all config/cache files are stored here.
ROOT = pathlib.Path.home() / ".local" / "share" / "tankobon"
ROOT.mkdir(parents=True, exist_ok=True)


def plural(n_items, noun):
    return f"{n_items} {noun}{'s' if n_items > 1 else ''}"


def filesize(content: bytes) -> str:
    """Create a human-readable filesize for content.

    Args:
        content: The bytes to get the size of.
    Returns:
        A string of the filesize ending in B, kB, etc.
    """
    filesize = float(len(content))
    for suffix in ["B", "KiB", "MiB", "GiB"]:
        if filesize < 1024.0 or suffix == "GiB":
            break
        filesize /= 1024.0
    return f"{filesize:.1f} {suffix}"


def _is_valid_char(char):
    return char.isalnum()


def sanitize(name: str) -> str:
    """Sanitise a name so it can be used as a filename.
    Args:
        name: The name to sanitise.
    Returns:
        The sanitised name as a string.
    """

    sanitised = "".join([c.lower() if _is_valid_char(c) else "_" for c in name])

    # remove duplicate underscores
    return re.sub("_{2,}", "_", sanitised).strip("_")


def soup(
    url: str, *args, session: Optional[requests.Session] = None, **kwargs
) -> bs4.BeautifulSoup:
    """Get a url as a BeautifulSoup.

    Args:
        url: The url to get a soup from.
        *args: Passed to session.get().
        session: The session to use to download the soup.
            Defaults to None.
        **kwargs: Passed to session.get().
    """

    if session is None:
        session = requests.Session()

    response = session.get(url, *args, **kwargs)

    return bs4.BeautifulSoup(response.text, BS4_PARSER)


def save_response(path: pathlib.Path, res: requests.models.Response) -> pathlib.Path:
    """Save a Requests response at path with the correct file extension.

    Args:
        path: The path where to save the file at.
        res: The response.

    Returns:
        The full path to the file.
    """

    path = path.with_suffix(f".{filetype.guess(res.content).extension}")
    with path.open("wb") as f:
        f.write(res.content)
    _log.debug(
        "saved response from %s at %s (%s)", res.url, path, filesize(res.content)
    )

    return path


def is_url(url: str) -> bool:
    """Check whether or not a string is a url."""
    result = urlparse(url)
    return all([result.scheme, result.netloc, result.path])


def parse_domain(url: str) -> str:
    """Parse out a url's domain.

    Args:
        url: The url to parse.

    Returns:
        The domain.
    """

    return RE_DOMAIN.findall(urlparse(url).netloc)[0][0]


class UserSession(requests.Session):
    """requests.Session with randomised user agent in the headers."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.headers.update({"User-Agent": USER_AGENT.random})


class PersistentDict(collections.UserDict):
    """A UserDict that can be loaded and dumped to disk persistently.
    (As long as the dictionary contents can be serialised to JSON.)

    Usage:

    ```python
    from tankobon.utils import PersistentDict

    file = "test.json"

    with PersistentDict(file) as d:
        d["foo"] = "bar"

    # '/where/to/save.json' now looks like this:
    # {
    #     "foo": "bar"
    # }

    # It can also be used without a context manager.
    # Just remember to close() it, or any changes won't be written to disk!

    d = PersistentDict(file)
    d["baz"] = 42
    d.close()
    ```
    """

    def __init__(
        self, path: Union[str, pathlib.Path], *args, compress: bool = False, **kwargs
    ):
        super().__init__(*args, **kwargs)

        if isinstance(path, str):
            path = pathlib.Path(path)

        self.path = path
        self.compress = compress

        try:
            if compress:
                f = gzip.open(self.path, "rt")
            else:
                f = self.path.open()

            self.data.update(json.load(f))

            f.close()

        except FileNotFoundError:
            pass

    def close(self):
        if self.compress:
            f = gzip.open(self.path, "wt")
        else:
            f = self.path.open("w")

        json.dump(self.data, f, indent=2)
        f.close()

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()


class Config(PersistentDict):

    CONFIG = "config.json"
    DEFAULTS = {"lang": "en"}

    def __init__(self, path: Optional[pathlib.Path] = None):
        if path is None:
            path = ROOT / self.CONFIG

        super().__init__(path, compress=False, **self.DEFAULTS)

    def __getitem__(self, key):
        if key not in self:
            super().__setitem__(key, os.environ[f"TANKOBON_{key.upper()}"])

        return super().__getitem__(key)



# coding: utf8
"""
# tankobon, version {version}

Copyright (c) 2020-2021 Ong Yong Xin

Licensed under the MIT License.

star this project at [ongyx/tankobon](https://github.com/ongyx/tankobon) or something, idk

sources:

{supported}
"""

import pathlib
import signal
import sys
import threading
import traceback

from PySide6.QtCore import Qt, Signal, QSize
from PySide6.QtGui import QAction, QIcon, QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QDialogButtonBox,
    QGridLayout,
    QHBoxLayout,
    QInputDialog,
    QMainWindow,
    QMessageBox,
    QMenuBar,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QProgressDialog,
    QScrollArea,
    QSizePolicy,
    QSplashScreen,
    QTextBrowser,
    QToolBar,
    QToolButton,
    QVBoxLayout,
    QWidget,
    QWidgetItem,
)

from .. import core, models, sources  # noqa: F401
from ..utils import Config
from . import resources, template, utils  # noqa: F401

from ..__version__ import __version__

_app = QApplication([])
QStyle = _app.style()

LOGO = QPixmap(":/logo.jpg")

HOME = pathlib.Path.home()

CONFIG = Config()
LANG = CONFIG["lang"]

CACHE = core.Cache()

MAX_COL = 5

T_ADD = "Add Manga"
T_DELETE = "Delete Manga"
T_DOWNLOAD = "Download Manga"

MANGA = {}
MANGA_LOCK = threading.Lock()


def _is_ascii(s):
    try:
        s.encode(encoding="utf8").decode("ascii")
    except UnicodeDecodeError:
        return False
    else:
        return True


def _normalize(s):
    return s.replace("_", " ").capitalize()


def _load_manga(hash):
    with MANGA_LOCK:
        if hash not in MANGA:
            MANGA[hash] = CACHE.load(hash)

        return MANGA[hash]


def delete(widget):

    if isinstance(widget, QWidgetItem):
        widget.widget().close()

    else:
        widget.hide()
        widget.deleteLater()


class SpinningCursor:
    def __enter__(self):
        _app.setOverrideCursor(Qt.WaitCursor)

    def __exit__(self, t, v, tb):
        _app.restoreOverrideCursor()


class TitleLabel(QLabel):
    def __init__(self, *args):
        super().__init__(*args)
        self.setTextFormat(Qt.RichText)
        self.setAlignment(Qt.AlignCenter)
        self.setWordWrap(True)
        self.setStyleSheet("background-color: #CCCCFF; color: black;")
        self.setAutoFillBackground(True)


class SubtitleLabel(QLabel):
    def __init__(self, subtitle):
        super().__init__(f"<b>{subtitle}</b>")
        self.setAlignment(Qt.AlignLeft | Qt.AlignTop)


# A message box without the window icon.
class MessageBox(QMessageBox):
    def __init__(self, *args):
        super().__init__(*args)

        self.setWindowIcon(QIcon(LOGO))

    @classmethod
    def info(cls, title, text):
        msgbox = cls(cls.Information, title, text, cls.Ok)
        return msgbox.exec()

    @classmethod
    def ask(cls, title, text):
        msgbox = cls(cls.Question, title, text, cls.Yes | cls.No)
        return msgbox.exec()

    @classmethod
    def warn(cls, title, text):
        msgbox = cls(cls.Warning, title, text, cls.Ok)
        return msgbox.exec()

    @classmethod
    def crit(cls, title, text):
        msgbox = cls(cls.Critical, title, text, cls.Ok)
        return msgbox.exec()


class AboutBox(MessageBox):
    def __init__(self, *args):
        super().__init__(*args)

        self.setWindowTitle("About")

        # build table of supported sources
        supported = []
        for cls in core.Parser.registered:
            supported.append(f"`{cls.__module__}` ({cls.domain.pattern})  ")

        self.setTextFormat(Qt.MarkdownText)
        self.setText(
            __doc__.format(
                version=__version__,
                supported="\n".join(supported),
            )
        )

        self.setAttribute(Qt.WA_DeleteOnClose)

        small_logo = LOGO.scaled(
            QSize(256, 256), Qt.KeepAspectRatio, Qt.SmoothTransformation
        )

        self.setIconPixmap(small_logo)


def _excepthook(ex_type, ex_value, ex_traceback):
    MessageBox.crit(
        "An exception occured.",
        "".join(traceback.format_exception(ex_type, ex_value, ex_traceback)),
    )


sys.excepthook = _excepthook


# A text dialog that requires input before allowing 'ok' to be pressed.
class RequiredDialog(QInputDialog):
    def __init__(self, *args):
        super().__init__(*args)

        self.setWindowIcon(QIcon(LOGO))

        self.textValueChanged.connect(self.onTextValueChanged)

        self.setInputMode(self.TextInput)

        self.setOkButtonText("Ok")

        self.ok_button, _ = self.findChild(QDialogButtonBox).buttons()
        self.ok_button.setEnabled(False)

    def onTextValueChanged(self, text):
        if text:
            self.ok_button.setEnabled(True)
        else:
            self.ok_button.setEnabled(False)


class ProgressDialog(QProgressDialog):
    def __init__(self, *args):
        super().__init__(*args)

        self.setMinimumDuration(0)
        self.setWindowModality(Qt.WindowModal)
        self.setAttribute(Qt.WA_DeleteOnClose, True)


# A manga item.
class Item(QListWidgetItem):
    def __init__(self, meta: models.Metadata):
        self.meta = meta
        super().__init__(self.meta.title)

        # self.setToolTip(", ".join(self.meta.alt_titles))


# A preview of the manga infomation (title, author, etc.)
class ItemInfoBox(QWidget):
    def __init__(self, item: Item):
        super().__init__()
        layout = QGridLayout(self)

        meta = item.meta

        # infobox spans one row and two columns.
        SPAN = (1, 2)

        # wikipedia-style info box at the side
        title = TitleLabel(f"<h2><i>{meta.title}</i></h2>")
        layout.addWidget(title, 0, 0, *SPAN)

        cover = QPixmap()
        try:
            manga_path = CACHE.root / meta.hash
            cover_path = next(manga_path.glob("cover.*"))

        except StopIteration:
            cover = QPixmap(":/missing.jpg")

        else:
            cover.load(str(cover_path))

        # scale cover
        self.cover = cover.scaled(
            self.width(),
            self.height(),
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation,
        )
        self.cover_label = QLabel()
        self.cover_label.setPixmap(self.cover)
        self.resizeCover()
        layout.addWidget(self.cover_label, 1, 0, *SPAN)

        if meta.alt_titles is not None:
            _alt_titles = "<br>".join(
                f"<i>{t}</i>" if _is_ascii(t) else t for t in meta.alt_titles
            )

        else:
            _alt_titles = "(empty)"

        alt_titles = TitleLabel(_alt_titles)
        alt_titles.setStyleSheet("background-color: #DDDDFF; color: black;")
        layout.addWidget(alt_titles, 2, 0, *SPAN)

        genre_header = SubtitleLabel("Genre")
        layout.addWidget(genre_header, 3, 0)

        if meta.genres is not None:
            _genres = "<br>".join(_normalize(g) for g in meta.genres)

        else:
            _genres = "(empty)"

        genres = QLabel(_genres)
        layout.addWidget(genres, 3, 1)

        manga_header = TitleLabel("<b>Manga</b>")
        layout.addWidget(manga_header, 4, 0, *SPAN)

        author_header = SubtitleLabel("Authored by")
        layout.addWidget(author_header, 5, 0)

        if meta.authors is not None:
            _authors = "<br>".join(a for a in meta.authors)

        else:
            _authors = "(empty)"

        authors = QLabel(_authors)
        layout.addWidget(authors, 5, 1)

        source_header = SubtitleLabel("Source")
        layout.addWidget(source_header, 6, 0)

        source = QLabel(f'<a href="{meta.url}">{meta.url}</b>')
        source.setWordWrap(True)
        layout.addWidget(source, 6, 1)

    def resizeCover(self):
        self.cover = self.cover.scaled(
            int(self.width() / 2),
            self.height(),
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation,
        )
        self.cover_label.setPixmap(self.cover)


# A list of manga items in the sidebar.
class ItemList(QListWidget):
    def __init__(self):
        super().__init__()
        self.hashs = set()

        for _, manga in CACHE.data.items():
            self.addItem(Item(manga["meta"]))

        self.reload()

    def addItem(self, item):
        self.hashs.add(item.meta.hash)
        super().addItem(item)

    def reload(self):
        self.setMaximumWidth(self.sizeHintForColumn(0) + 5)


MANGA_ITEMS = ItemList()


# Toolbar at the bottom of the window.
# This shows a bunch of buttons to manage manga items (add, remove, etc.)
class ToolBar(QToolBar):

    deletedManga = Signal()

    BUTTONS = [
        ("add", ":/plus.svg"),
        (
            "delete",
            ":/minus.svg",
        ),
        (
            "refresh",
            ":/refresh-cw.svg",
        ),
        (
            "download",
            ":/download.svg",
        ),
        ("view", ":/eye.svg"),
        ("locate", ":/folder.svg"),
    ]

    def __init__(self):
        super().__init__()
        MANGA_ITEMS.itemClicked.connect(self.onSelectedManga)
        self.deletedManga.connect(self.onDeletedManga)

        self.selected = None

        self.summaries = {}

        bg_is_dark = utils.is_dark(_app.palette().window().color())

        for method_name, icon_path in self.BUTTONS:

            method = getattr(self, method_name)
            tooltip = f"{method_name.title()} a manga..."

            if bg_is_dark:
                icon = QIcon(icon_path.replace(".svg", "-light.svg"))
            else:
                icon = QIcon(icon_path)

            action = QAction()
            action.setToolTip(tooltip)
            action.setIcon(icon)
            action.triggered.connect(method)

            button = QToolButton()
            button.setDefaultAction(action)

            self.addWidget(button)

        # spacer
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.addWidget(spacer)

        # text below the manga preview
        self.summary = QLabel()
        self.addWidget(self.summary)

        self.show()

    def onSelectedManga(self, manga_item):
        self.selected = manga_item

        hash = manga_item.meta.hash
        if hash not in self.summaries:
            manga = _load_manga(hash)
            self.summaries[hash] = f"{len(manga.chapters)} chapters"

        self.summary.setText(self.summaries[hash])

    def onDeletedManga(self):
        self.summary.setText("")

    def ensureSelected(self, method):
        if self.selected is None:
            MessageBox.info(
                f"{method.capitalize()} Manga",
                f"Please select a manga to {method} first.",
            )
            return False

        return True

    def _refresh(self, manga):

        with SpinningCursor():
            parser = core.Parser.parser(manga.meta.url)
            parser.add_chapters(manga)

            CACHE.dump(manga)

            # add to item list (only if manga is new)
            if manga.meta.hash not in MANGA_ITEMS.hashs:
                MANGA_ITEMS.addItem(Item(manga.meta))
                MANGA_ITEMS.reload()

    def add(self):
        dialog = RequiredDialog()
        dialog.setWindowTitle(T_ADD)
        dialog.setLabelText("Enter the manga url below:")

        dialog_code = dialog.exec()
        if dialog_code == QInputDialog.Rejected:
            # canceled
            return

        url = dialog.textValue()

        if url in CACHE.alias:
            MessageBox.info(
                T_ADD,
                "Manga already exists in cache. To refresh a manga, select a manga and click the refresh button.",
            )
            return

        try:
            parser = core.Parser.parser(url)
        except core.UnknownDomainError:
            MessageBox.warn(
                T_ADD,
                "Manga url is invalid or no parser was found for the url.",
            )
            return

        manga = parser.create(url)

        self._refresh(manga)

        if manga.meta.cover:
            with core.Downloader(CACHE.root / manga.meta.hash) as downloader:
                downloader.download_cover(manga)

    def delete(self):
        if not self.ensureSelected("delete"):
            return

        reply = MessageBox.ask(
            T_DELETE,
            "Are you sure you want to delete this manga? This cannot be undone!",
        )

        if reply == MessageBox.Yes:
            CACHE.delete(self.selected.meta.hash)
            MANGA_ITEMS.takeItem(MANGA_ITEMS.row(self.selected))
            MANGA_ITEMS.reload()

            self.deletedManga.emit()

    def refresh(self):
        if not self.ensureSelected("refresh"):
            return

        manga = _load_manga(self.selected.meta.hash)
        self._refresh(manga)

    def _download(self, manga, chapters):
        dialog = ProgressDialog(self)

        parser = core.Parser.parser(manga.meta.url)

        with core.Downloader(CACHE.root / manga.meta.hash) as downloader:
            for count, chapter in enumerate(chapters):

                dialog.setLabelText(f"Downloading chapter {chapter.id}...")

                if not chapter.pages:
                    parser.add_pages(chapter)

                total = len(chapter.pages)

                # HACK: make the progress bar display properly during first iteration
                dialog.setMaximum(total)
                dialog.setValue(total - 1)
                dialog.setValue(0)

                if dialog.wasCanceled():
                    break

                downloader.download(chapter, progress=dialog.setValue)

                dialog.setValue(total)

    def download(self):
        if not self.ensureSelected("download"):
            return

        manga = _load_manga(self.selected.meta.hash)

        dialog = RequiredDialog()
        dialog.setWindowTitle(T_DOWNLOAD)
        dialog.setLabelText(
            "Enter the chapters to download below, seperated by commas.\n"
            "Ranges are also allowed, i.e 1-5."
        )

        dialog_code = dialog.exec()
        if dialog_code == QInputDialog.Rejected:
            return

        chapters = manga.select(dialog.textValue(), lang=LANG)
        if not chapters:
            MessageBox.warn(T_DOWNLOAD, "Chapters/range is invalid.")

        self._download(manga, chapters)

    def view(self):
        if not self.ensureSelected("view"):
            return

        MessageBox.info("oops", "Sorry! The viewer isn't implemented yet.")

    def locate(self):
        if not self.ensureSelected("locate"):
            return

        hash = self.selected.meta.hash

        utils.xopen(str(CACHE.root / hash))


# Toolbar at the top of the window.
class MenuBar(QMenuBar):
    def __init__(self):
        super().__init__()

        file = self.addMenu("File")

        file_quit = QAction("Quit", self)
        file_quit.triggered.connect(_app.quit)

        file.addAction(file_quit)

        help = self.addMenu("Help")

        help_tankobon = QAction("About tankobon", self)
        help_tankobon.triggered.connect(self.about)

        help.addAction(help_tankobon)

        help_qt = QAction("About Qt", self)
        help_qt.triggered.connect(_app.aboutQt)

        help.addAction(help_qt)

    def about(self):
        about_box = AboutBox(self)
        about_box.exec()


# The combined manga item list plus preview.
class View(QWidget):
    def __init__(self):
        super().__init__()
        self.layout = QHBoxLayout(self)

        self.pixmap_cache = {}

        # The split view at first shows the list of manga items and the default item view.
        # After a manga has been selected, there will be a total of three widgets:
        # - Manga item list
        # - Manga info.
        # - Manga cover
        self.layout.addWidget(MANGA_ITEMS)
        self.layout.addWidget(self.default())

        MANGA_ITEMS.itemClicked.connect(self.onSelectedManga)

    def default(self):
        label = QLabel("Select a manga from the side bar, or add a new one.")
        label.setAlignment(Qt.AlignCenter)

        return label

    def deleteLast(self):
        while self.layout.count() != 1:
            delete(self.layout.takeAt(1))

    def onSelectedManga(self, manga_item):
        self.deleteLast()

        manga = _load_manga(manga_item.meta.hash)

        text = QTextBrowser()
        text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        text.setReadOnly(True)
        text.setOpenExternalLinks(True)
        text.document().setDefaultStyleSheet(
            utils.resource(":/view.css").decode("utf8")
        )
        text.setHtml(template.create(manga))

        self.layout.addWidget(text)

        infobox = ItemInfoBox(manga_item)
        scroll = QScrollArea()
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.verticalScrollBar().setStyleSheet("height:0px;")
        scroll.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        scroll.setWidget(infobox)

        self.layout.addWidget(scroll)

    def onDeletedManga(self):
        self.deleteLast()

        self.layout.addWidget(self.default())


# Main window.
class Root(QMainWindow):
    def __init__(self):
        super().__init__()

        # gui setup
        self.setWindowTitle("tankobon")
        self.setWindowIcon(LOGO)

        # can't direcly apply layouts to the main window, so wrap in a widget.
        wrapper = QWidget(self)
        self.layout = QVBoxLayout(wrapper)

        self.menubar = MenuBar()
        self.setMenuBar(self.menubar)

        self.view = View()
        self.layout.addWidget(self.view)

        self.toolbar = ToolBar()
        self.layout.addWidget(self.toolbar)

        self.toolbar.deletedManga.connect(self.view.onDeletedManga)

        self.setCentralWidget(wrapper)

    def confirmQuit(self):
        reply = MessageBox.ask(
            "Quit?",
            "Are you sure you want to exit?",
        )
        return reply == MessageBox.Yes

    def closeEvent(self, event):
        if self.confirmQuit():

            CACHE.close()
            event.accept()
        else:
            event.ignore()


class LoadingSplash(QSplashScreen):
    def showMessage(self, message):
        super().showMessage(
            f"\ntankobon v{__version__}\nCopyright (c) 2020-2021 Ong Yong Xin\n{message}",
            Qt.AlignBottom,
            Qt.white,
        )


def main():
    splash = LoadingSplash(LOGO)
    splash.show()

    splash.showMessage("Loading manga metadata...")

    window = Root()

    window.show()
    splash.finish(window)

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    sys.exit(_app.exec_())



# coding: utf8

import json

from tankobon.jsonclasses import dataclass  # type: ignore


@dataclass
class Case:
    foo: int
    bar: str = "baz"


@dataclass
class NestedCase:
    foo: Case


def test_dumps_and_loads():

    instance = Case(0)

    dumped = json.dumps(instance, indent=4)
    print(dumped)

    loaded = json.loads(dumped)
    print(loaded)

    assert isinstance(loaded, Case)


def test_nested():
    instance = NestedCase(Case(0))

    dumped = json.dumps(instance)
    print(dumped)



# coding: utf8
"""This directory contains sources for several websites.

Sources are regular modules that have the following classes:

## `Parser`

Subclass of `tankobon.core.Parser`.

`Parser`s must implement the methods `metadata`, `chapters` and `pages`.

It must also have a class attribute `domain`, which is a uncompiled regex pattern of the urls the parser can parse.
At minimum it should be the base of the url (**without** http(s):// or www. in front):

```python
domain = r"my-manga-host.com"
```

When subclassed, the new parser will automatically be registered (as long as it runs, i.e import it).
The parser will then be delegated to based on the domain.

For some examples, take a look at the sources in this directory.
"""

# import to register default sources
from . import catmanga, mangadex, mangakakalot



# coding: utf8


class TankobonError(Exception):
    pass


class MangaError(TankobonError):
    pass


class MangaNotFoundError(MangaError):
    pass


class PagesNotFoundError(MangaError):
    pass


class UnknownDomainError(TankobonError):
    pass



# coding: utf8
"""This template generates the HTML view for a manga."""

import string

from PySide6.QtWidgets import QTextEdit

from ..models import Manga

from .utils import resource

TEMPLATE = string.Template(resource(":/view.html").decode("utf8"))


def create(manga: Manga):
    # convert table to html from markdown
    textedit = QTextEdit()
    textedit.setMarkdown(manga.summary())

    table = textedit.toHtml()

    return TEMPLATE.substitute(
        url=manga.meta.url, title=manga.meta.title, desc=manga.meta.desc, table=table
    )



# coding: utf8

import re

from .. import core, models


# mangakakalot chapter urls always end with '/chapter_(number)'.
RE_CHAPTER = re.compile(r"(\d+(\.\d+)?)")
RE_TITLE = re.compile(r"^.*: (.*)$")


class Parser(core.Parser):

    domain = r"mangakakalot.com"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._url = None
        self._soup = None

    def _get_soup(self, url):
        if url != self._url:
            self._soup = self.soup(url)

        return self._soup

    def metadata(self, url):
        soup = self._get_soup(url)
        info = soup.find(class_="manga-info-text").find_all("li")

        title_tag = info[0]
        title = title_tag.h1.text

        alt_titles_tag = title_tag.h2.text.partition(":")[-1]
        alt_titles = alt_titles_tag.split(",")

        authors = [a.text for a in info[1].find_all("a")]

        genres = [a.text.lower().replace(" ", "_") for a in info[6].find_all("a")]

        desc_tag = soup.find("div", id=["panel-story-info-description", "noidungm"])

        try:
            desc_tag.p.decompose()
        except AttributeError:
            pass
        finally:
            desc = desc_tag.text

        cover = soup.find("div", class_="manga-info-pic").img["src"]
        print(cover)

        return models.Metadata(
            url=url,
            title=title,
            alt_titles=alt_titles,
            authors=authors,
            genres=genres,
            desc=desc,
            cover=cover,
        )

    def add_chapters(self, manga):
        soup = self._get_soup(manga.meta.url)

        for tag in soup.find_all("div", class_="row"):
            link = tag.span.a
            if not link:
                continue

            url = link["href"]
            cid = RE_CHAPTER.search(url.split("/")[-1]).group(1)
            title = RE_TITLE.match(link["title"]).group(1)

            manga.add(
                models.Chapter(
                    id=cid,
                    url=url,
                    title=title,
                )
            )

    def add_pages(self, chapter):
        soup = self.soup(chapter.url)
        chapter.pages = [
            tag["src"]
            for tag in soup.find("div", class_="container-chapter-reader").find_all(
                "img", src=True
            )
        ]



# coding: utf8
"""Model classes."""

from __future__ import annotations

import collections
import hashlib
import logging
from typing import cast, Any, Dict, List, Optional

import natsort  # type: ignore

from . import utils
from .jsonclasses import dataclass, field


# cast because mypy keeps complaining about generic types
def _list():
    return cast(List[Any], field(default_factory=list))


def _dict():
    return cast(Dict[str, Any], field(default_factory=dict))


_log = logging.getLogger("tankobon")


@dataclass
class Metadata:
    """Metadata for a manga.

    Args:
        url: The url to the manga title page.
        title: The manga name in English (romanized/translated).
        alt_titles: A list of alternative names for the manga.
            i.e in another language, original Japanese name, etc.
        desc: The sypnosis (human-readable info) of the manga.
        cover: The url to the manga cover page (must be an image).
        authors: A list of author names.
        genres: A list of catagories the manga belongs to.
            i.e shounen, slice_of_life, etc.
            Note that the catagories are sanitised using utils.sanitise() on initalisation.
        other: Miscellanious map of keys to values.
            May be used by parsers to store parser-specific info (keep state).

    Attributes:
        hash: A SHA-256 checksum of the manga url.
            (Can be used for filename-safe manga storage.)
    """

    url: str
    title: str = ""

    alt_titles: List[str] = _list()

    desc: str = ""
    cover: str = ""

    authors: List[str] = _list()
    genres: List[str] = _list()

    hash: str = ""

    other: Dict[str, Any] = _dict()

    def __post_init__(self):
        if self.genres:
            self.genres = [utils.sanitize(g.strip()) for g in self.genres]

        self.desc = self.desc.strip().replace("\r\n", "\n")

        if not self.hash:
            self.hash = hashlib.sha256(self.url.encode()).hexdigest()


@dataclass
class Chapter:
    """A manga chapter.

    Args:
        id: The chapter id as a string (i.e 1, 2, 10a, etc.).
        url: The chapter url.
        title: The chapter name.
        volume: The volume the chapter belongs to.
        lang: The ISO 639-1 language code that this chapter was translated to.
        pages: A list of image urls to the chapter pages.
        other: Miscellanious map of keys to values.
            May be used by parsers to store parser-specific info (keep state).
    """

    id: str
    url: str
    title: str = ""
    volume: str = "0"
    lang: str = "en"

    pages: List[str] = _list()

    other: Dict[str, Any] = _dict()


class Manga:
    """A manga.

    Selecting chapters in this manga can be done by slicing:

    manga[start_cid:end_cid:lang]  # returns a list of Chapter objects

    where start_cid is the first chapter of the selection, and end_cid is the last chapter of the selection.
    lang is the ISO 639-1 language code of the chapters to select. i.e:

    # Select chapters 1 to 5 in the Spanish language (inclusive of chapter 5).
    # NOTE: If the chapter does not have a translation for the selected language,
    # the number of chapters you get may not be the number requested!
    chapters = manga["1":"5":"es"]

    Args:
        meta: The manga metadata.
        chapters: The manga chapters.

    Attributes:
        chapters: A map of chapter ids to a map of ISO 639-1 language codes to Chapter objects (chapters may have several languages):

            {
                // chapter id
                "1": {
                    // ISO 639-1 language code
                    "en": Chapter(...)
                }
            }

        info: A dictionary which has the following keys:

            chapters (int)
                The total number of chapters across all languages.

            volumes (set)
                The volumes that this manga has across all languages.

            langs (set)
                ISO 639-1 language codes that this manga was translated to.
                Note that chapters may not have a translation for all language codes.
    """

    def __init__(
        self,
        metadata: Metadata,
        chapters: Optional[dict] = None,
    ):
        self.meta = metadata

        self.chapters: Dict[str, Dict[str, Chapter]]

        self.chapters = collections.defaultdict(dict)

        if chapters is not None:
            self.chapters.update(chapters)

    @property
    def info(self):
        info = {
            "chapters": 0,
            "volumes": set(),
            "langs": set(),
        }

        for cid, langs in self.chapters.items():
            info["langs"].update(langs.keys())

            for chapter in langs.values():
                info["chapters"] += 1
                info["volumes"].add(chapter.volume)

        return info

    def add(self, chapter: Chapter):
        """Add a chapter to this manga.
        The chapter will not be added if it already exists (has the same id and lang as the existing one).

        Args:
            chapter: The chapter to add.
        """

        if not self.exists(chapter):
            self.chapters[chapter.id][chapter.lang] = chapter

    def remove(self, cid: str, lang: str = "en") -> Chapter:
        """Remove a chapter from this manga.

        Args:
            cid: The chapter id to remove.
            lang: The chapter language to remove.
                Defaults to 'en'.

        Returns:
            The removed chapter.
        """

        return self.chapters[cid].pop(lang)

    def exists(self, chapter: Chapter) -> bool:
        """Check whether a chapter already exists in this manga.

        Args:
            chapter: The chapter object.

        Returns:
            True if it exists, otherwise False.
        """

        langs = self.chapters.get(chapter.id, {})

        return bool(langs) and chapter.lang in langs

    def dump(self) -> dict:
        """Serialise this manga to a dict."""
        return {"meta": self.meta, "chapters": self.chapters}

    @classmethod
    def load(cls, data: dict) -> Manga:
        """Deserialise this manga from a dict.

        Args:
            data: The serialised manga.

        Returns:
            The Manga object.
        """
        return cls(data["meta"], chapters=data["chapters"])  # type: ignore

    def summary(self, lang: str = "en") -> str:
        """Create a Markdown table summary of all volumes and chapters in this manga.

        Args:
            lang: The language to summerise for.

        Returns:
            The Markdown table as a string.
        """

        table = ["| volume | chapter | title ", "|--------|---------|-------"]

        for cid, langs in natsort.natsorted(self.chapters.items()):

            chapter: Chapter = langs.get(lang)

            if chapter is None:
                # chapter does not have the language.
                continue

            table.append(
                "| {:<6} | {:<7} | [{}]({})".format(
                    chapter.volume or "(empty)",
                    chapter.id or "(empty)",
                    chapter.title or "(empty)",
                    chapter.url,
                )
            )

        return "\n".join(table)

    def select(self, cids: str, lang: str = "en") -> List[Chapter]:
        """Select chapters from this manga.

        Args:
            cids: A list of chapter ids as a string, delimited by a comma.
                Ranges are also valid (1-5).
                i.e '1,3,5,8-10' (select chapters 1,3,5 and 8-10 inclusive of 10).
            lang: The language of the chapters.
                Note that if a chapter does not have the language requested, it will be skipped.

        Returns:
            A list of Chapter objects.
        """

        chapters = []

        for cid in cids.split(","):

            if "-" in cid:
                start, end = cid.split("-")
                chapters.extend(self[start:end:lang])  # type: ignore

            else:
                langs = self.chapters[cid]
                chapter = langs.get(lang)

                if chapter is not None:
                    chapters.append(chapter)

        return chapters

    def parsed(self) -> bool:
        """Check whether this manga has been parsed (has at least one chapter)."""
        return len(self.chapters) >= 1

    def __getitem__(self, key):
        if isinstance(key, slice):
            sorted_chapters = natsort.natsorted(self.chapters.keys())

            try:
                start = sorted_chapters.index(key.start)
                if key.stop is not None:
                    # include the last chapter also
                    stop = sorted_chapters.index(key.stop, start) + 1
                else:
                    stop = len(sorted_chapters)
            except ValueError:
                return []

            lang = "en" if key.step is None else key.step

            return [
                self.chapters[c][lang]
                for c in sorted_chapters[start:stop]
                if lang in self.chapters[c]
            ]

        else:
            return self.chapters[key]

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()



# coding: utf8
"""Wraps the json module to allow (de)serializing dataclasses.
(As long as your dataclass uses only the basic JSON types, no nested classes.)

To use, just replace any `dataclasses` import with `jsonclasses`.
jsonclasses automatically monkeypatches the json module so that it can dump any dataclasses created using the `@jsonclasses.dataclass` decorator.

NOTE: Any dataclass fields that start with `_` (i.e `_data`) **will** be treated as private and will not be serialised.
"""

import functools
import json
from dataclasses import (
    dataclass as _dataclass,
    field,
    Field,
    FrozenInstanceError,
    InitVar,
    MISSING,
    fields,
    asdict,
    astuple,
    make_dataclass,
    replace,
    is_dataclass,
)
from typing import Any, Dict

__all__ = [
    "dataclass",
    "field",
    "Field",
    "FrozenInstanceError",
    "InitVar",
    "MISSING",
    "fields",
    "asdict",
    "astuple",
    "make_dataclass",
    "replace",
    "is_dataclass",
]


_DATACLASSES: Dict[str, Any] = {}


def dataclass(cls=None, **kwargs):
    def wrap(cls):
        _DATACLASSES[cls.__name__] = cls
        return _dataclass(cls, **kwargs)

    if cls is None:
        return wrap

    return wrap(cls)


def _encode(o):
    if is_dataclass(o):
        new_o = {k: v for k, v in o.__dict__.items() if not k.startswith("_")}
        new_o["__dataclass__"] = o.__class__.__name__
        return new_o

    return json.JSONEncoder().default(o)


def _decode(o):
    if "__dataclass__" in o:
        cls = _DATACLASSES[o.pop("__dataclass__")]
        return cls(**o)
    return o


# monkey-patch load(s) and dump(s)
json.loads = functools.partial(json.loads, object_hook=_decode)
json.load = functools.partial(json.load, object_hook=_decode)
json.dumps = functools.partial(json.dumps, default=_encode)
json.dump = functools.partial(json.dump, default=_encode)



# coding: utf8

import functools
import logging
import pathlib
from typing import Any, Callable

import click
import coloredlogs  # type: ignore

from . import __version__, core, sources, utils  # noqa: F401
from .exceptions import MangaNotFoundError

click.option: Callable[..., Any] = functools.partial(click.option, show_default=True)  # type: ignore

_log = logging.getLogger("tankobon")
VERBOSITY = [
    getattr(logging, level)
    for level in ("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG")
]

CONFIG = utils.Config()
LANG = CONFIG["lang"]


@click.group()
@click.version_option(__version__)
@click.option(
    "-v", "--verbose", "verbosity", help="be more chatty", default=4, count=True
)
def cli(verbosity):
    """Manga browser/downloader.

    Once a manga has been added to tankobon using 'tankobon add <url>',
    you can refer to it using its shorthash (first 8 characters of its SHA512 hash).

    Executing 'tankobon list' will show the shorthashes of all manga added to tankobon.
    """
    # set up logger
    coloredlogs.install(
        level=VERBOSITY[verbosity - 1],
        fmt=" %(levelname)-8s :: %(message)s",
        logger=_log,
    )


def prettyprint(dict_):

    for key, value in dict_.items():

        # +2 for the ': ' suffix
        indent = len(key) + 2

        if isinstance(value, list):
            value = f"\n{' ' * indent}".join(value)

        click.echo(f"{key}: {value}\n")


def _load(shorthash, cache):
    try:
        return cache.load(cache.fullhash(shorthash))
    except MangaNotFoundError:
        click.echo(
            f"Manga not found in the cache. Try adding it first with 'tankobon add {shorthash}'."
        )
        raise click.Abort()


@cli.command()
@click.argument("shorthash")
@click.option("-c", "--chapter", help="show info only for a specific chapter")
def info(shorthash, chapter):
    """Show info on a manga."""

    with core.Cache() as cache:

        manga = _load(shorthash, cache)

        if chapter:
            prettyprint(manga.chapters[chapter].__dict__)
        else:
            prettyprint(manga.meta.__dict__)

            click.echo(manga.summary())

            info = manga.info

            click.echo(
                "summary: "
                f"{utils.plural(len(info['volumes']), 'volume')}, "
                f"{utils.plural(info['chapters'], 'chapter')}\n"
                f"languages: {', '.join(info['langs'])}"
            )


@cli.command("list")
def _list():
    """List all manga in the cache."""

    prettyprint(
        {"supported websites": [cls.domain.pattern for cls in core.Parser.registered]}
    )

    with core.Cache() as cache:

        if not cache.data:
            click.echo("(none)")

        else:
            for hash, manga in cache.data.items():
                meta = manga["meta"]

                print(f"{hash[:core.SHORT_HASH_LEN]}: {meta.title} ({meta.url})")


def _refresh(manga):
    parser = core.Parser.parser(manga.meta.url)
    parser.add_chapters(manga)


@cli.command()
@click.argument("url")
def add(url):
    """Create a manga by url."""

    with core.Cache() as cache:

        if url in cache.alias:
            short_hash = cache.alias[url][: core.SHORT_HASH_LEN]

            click.echo(
                f"manga already exists (refresh with 'tankobon refresh {short_hash}')"
            )
            return

        parser = core.Parser.parser(url)
        manga = parser.create(url)

        _refresh(manga)

        cache.dump(manga)

        click.echo(
            f"Sucessfully added '{url}'! It's shorthash is <{manga.meta.hash[:core.SHORT_HASH_LEN]}>."
        )


@cli.command()
@click.argument("shorthash")
def refresh(shorthash):
    """Refresh a manga by shorthash (adds any new chapters)."""

    with core.Cache() as cache:

        manga = _load(shorthash, cache)
        _refresh(manga)
        cache.dump(manga)


@cli.command()
@click.argument("shorthash")
def remove(shorthash):
    """Delete a manga by shorthash from the cache."""

    with core.Cache() as cache:
        cache.delete(cache.fullhash(shorthash))


@cli.command()
@click.argument("shorthash")
@click.option(
    "-p",
    "--path",
    type=pathlib.Path,
    default=".",
    help="where to download to (must be a folder)",
)
@click.option(
    "-c",
    "--chapters",
    "cids",
    help=(
        "chapters to download, seperated by ','. "
        "Ranges are also allowed (i.e '1-5,10' - download chapters 1 to 5 and 10)."
    ),
)
@click.option("-f", "--force", help="redownload existing chapters", default=False)
def download(shorthash, path, cids, force):
    """Download a manga by shorthash."""

    cache = core.Cache()
    parser = core.Parser.parser(shorthash)
    downloader = core.Downloader(path)

    manga = _load(shorthash, cache)

    if cids is None:
        if not click.confirm(
            "ALL chapters will be downloaded (this consumes a lot of bandwidth). Are you sure?"
        ):
            raise click.Abort()

        chapters = []

        for _, langs in manga.chapters.items():
            chapter = langs.get(LANG)
            if chapter is not None:
                chapters.append(chapter)

    else:
        chapters = manga.select(cids, lang=LANG)

    for chapter in chapters:
        click.echo(f"downloading chapter {chapter.id}")

        if not chapter.pages:
            click.echo(f"chapter {chapter.id} does not have any pages, adding")
            parser.add_pages(chapter)

        downloader.download(chapter, force=force)

    cache.close()
    downloader.close()


@cli.command()
@click.option(
    "-p",
    "--path",
    default=".",
    help="manga folder to make a pdf for",
)
@click.option(
    "-c", "--chapters", help="only add specific chapters to the pdf, split by ','"
)
@click.option(
    "-o",
    "--output",
    default="export.pdf",
    help="where to save the pdf",
)
def pdfify(path, chapters, output):
    """Create a (single) pdf file for all chapters of a downloaded manga."""

    with core.Downloader(path) as downloader:
        if chapters is None:
            chapters = list(downloader.manifest.keys())
        else:
            chapters = chapters.split(",")

        downloader.pdfify(chapters, output, lang=LANG)


@cli.command("gui")
def _gui():
    """Start the tankobon GUI."""

    try:
        from .gui import gui

    except ImportError:
        click.echo(
            "GUI extension not installed, install with 'pip install tankobon[gui]'."
        )
        raise click.Abort()

    else:
        gui.main()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['tankobon', 'tankobon.gui', 'tankobon.sources']

package_data = \
{'': ['*'],
 'tankobon': ['.mypy_cache/*',
              '.mypy_cache/3.9/*',
              '.mypy_cache/3.9/PySide6/*',
              '.mypy_cache/3.9/_typeshed/*',
              '.mypy_cache/3.9/click/*',
              '.mypy_cache/3.9/collections/*',
              '.mypy_cache/3.9/concurrent/*',
              '.mypy_cache/3.9/concurrent/futures/*',
              '.mypy_cache/3.9/ctypes/*',
              '.mypy_cache/3.9/email/*',
              '.mypy_cache/3.9/http/*',
              '.mypy_cache/3.9/importlib/*',
              '.mypy_cache/3.9/json/*',
              '.mypy_cache/3.9/logging/*',
              '.mypy_cache/3.9/multiprocessing/*',
              '.mypy_cache/3.9/os/*',
              '.mypy_cache/3.9/requests/*',
              '.mypy_cache/3.9/requests/packages/*',
              '.mypy_cache/3.9/requests/packages/urllib3/*',
              '.mypy_cache/3.9/requests/packages/urllib3/packages/*',
              '.mypy_cache/3.9/requests/packages/urllib3/packages/ssl_match_hostname/*',
              '.mypy_cache/3.9/requests/packages/urllib3/util/*',
              '.mypy_cache/3.9/tankobon/*',
              '.mypy_cache/3.9/tankobon/gui/*',
              '.mypy_cache/3.9/tankobon/parsers/*',
              '.mypy_cache/3.9/tankobon/sources/*',
              '.mypy_cache/3.9/urllib/*'],
 'tankobon.gui': ['.mypy_cache/*',
                  '.mypy_cache/3.9/*',
                  '.mypy_cache/3.9/PySide6/*',
                  '.mypy_cache/3.9/_typeshed/*',
                  '.mypy_cache/3.9/click/*',
                  '.mypy_cache/3.9/collections/*',
                  '.mypy_cache/3.9/concurrent/*',
                  '.mypy_cache/3.9/concurrent/futures/*',
                  '.mypy_cache/3.9/ctypes/*',
                  '.mypy_cache/3.9/email/*',
                  '.mypy_cache/3.9/http/*',
                  '.mypy_cache/3.9/importlib/*',
                  '.mypy_cache/3.9/json/*',
                  '.mypy_cache/3.9/logging/*',
                  '.mypy_cache/3.9/multiprocessing/*',
                  '.mypy_cache/3.9/os/*',
                  '.mypy_cache/3.9/requests/*',
                  '.mypy_cache/3.9/requests/packages/*',
                  '.mypy_cache/3.9/requests/packages/urllib3/*',
                  '.mypy_cache/3.9/requests/packages/urllib3/packages/*',
                  '.mypy_cache/3.9/requests/packages/urllib3/packages/ssl_match_hostname/*',
                  '.mypy_cache/3.9/requests/packages/urllib3/util/*',
                  '.mypy_cache/3.9/tankobon/*',
                  '.mypy_cache/3.9/tankobon/gui/*',
                  '.mypy_cache/3.9/tankobon/parsers/*',
                  '.mypy_cache/3.9/tankobon/sources/*',
                  '.mypy_cache/3.9/tankobon/tankobon/*',
                  '.mypy_cache/3.9/tankobon/tankobon/gui/*',
                  '.mypy_cache/3.9/urllib/*'],
 'tankobon.sources': ['.mypy_cache/*',
                      '.mypy_cache/3.9/*',
                      '.mypy_cache/3.9/PySide6/*',
                      '.mypy_cache/3.9/_typeshed/*',
                      '.mypy_cache/3.9/collections/*',
                      '.mypy_cache/3.9/concurrent/*',
                      '.mypy_cache/3.9/concurrent/futures/*',
                      '.mypy_cache/3.9/ctypes/*',
                      '.mypy_cache/3.9/email/*',
                      '.mypy_cache/3.9/http/*',
                      '.mypy_cache/3.9/importlib/*',
                      '.mypy_cache/3.9/json/*',
                      '.mypy_cache/3.9/logging/*',
                      '.mypy_cache/3.9/multiprocessing/*',
                      '.mypy_cache/3.9/os/*',
                      '.mypy_cache/3.9/requests/*',
                      '.mypy_cache/3.9/requests/packages/*',
                      '.mypy_cache/3.9/requests/packages/urllib3/*',
                      '.mypy_cache/3.9/requests/packages/urllib3/packages/*',
                      '.mypy_cache/3.9/requests/packages/urllib3/packages/ssl_match_hostname/*',
                      '.mypy_cache/3.9/requests/packages/urllib3/util/*',
                      '.mypy_cache/3.9/tankobon/*',
                      '.mypy_cache/3.9/tankobon/parsers/*',
                      '.mypy_cache/3.9/tankobon/sources/*',
                      '.mypy_cache/3.9/tankobon/tankobon/*',
                      '.mypy_cache/3.9/tankobon/tankobon/sources/*',
                      '.mypy_cache/3.9/urllib/*']}

install_requires = \
['beautifulsoup4>=4.9.1',
 'click>=7.1.2',
 'coloredlogs>=14.0',
 'dataclasses-json>=0.5.4',
 'fake-useragent>=0.1.11',
 'filetype>=1.0.7',
 'fpdf>=1.7.2',
 'html5lib>=1.1',
 'imagesize>=1.2.0',
 'MangaDex.py>=2.0.3',
 'natsort>=7.1.0',
 'requests>=2.24.0']

extras_require = \
{'dev': ['pytest', 'pytest-flake8', 'pytest-mypy', 'pydoc-markdown>=3.10.1'],
 'gui': ['PySide6>=6.0.3']}

entry_points = \
{'console_scripts': ['tankobon = tankobon.cli:cli']}

setup(name='tankobon',
      version='2021.6.3',
      description='Yet another manga scraper and downloader',
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/tankobon',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



# coding: utf8
"""Core functionality of tankobon."""

from __future__ import annotations

import abc
import concurrent.futures as cfutures
import gzip
import logging
import pathlib
import re
import shutil
from typing import cast, Callable, Dict, List, Optional, Type, Union

import bs4  # type: ignore
import fpdf  # type: ignore
import imagesize  # type: ignore
import natsort  # type: ignore
import requests.exceptions

from . import models, utils
from .exceptions import MangaNotFoundError, PagesNotFoundError, UnknownDomainError

_log = logging.getLogger("tankobon")

A4_WIDTH = 210
A4_HEIGHT = 297

SHORT_HASH_LEN = 8


class Parser(abc.ABC):

    # This allows subclasses to be registered.
    registered: List[Type[Parser]] = []

    def __init__(self):
        self.session = utils.UserSession()

    def create(self, url: str) -> models.Manga:
        """Create a new manga.

        Args:
            url: The manga url.

        Returns:
            A Manga object.
        """

        metadata = self.metadata(url)
        return models.Manga(metadata)

    @classmethod
    def parser(cls, url: str) -> Parser:
        """Get the appropiate parser subclass for the domain in url.

        Args:
            url: The url to get the subclass for.

        Returns:
            The subclass instance that can be used to parse the url.

        Raises:
            UnknownDomainError, if there is no registered subclass for the url domain.
        """

        for subclass in cls.registered:
            if subclass.domain.search(url):  # type: ignore
                return subclass()

        raise UnknownDomainError(f"no source found for url '{url}'")

    @property
    @abc.abstractmethod
    def domain(self) -> str:
        pass

    @abc.abstractmethod
    def metadata(self, url: str) -> models.Metadata:
        """Parse metadata for a manga url.

        Args:
            url: The manga url.

        Returns:
            The Metadata object.
        """

    @abc.abstractmethod
    def add_chapters(self, manga: models.Manga):
        """Add chapters to the manga.

        This method should add every chapter in the manga as a Chapter object:

        ```python
        def chapters(self, manga):
            for ... in ...:
                # do your parsing here
                manga.add(Chapter(...))
        ```

        Only the 'url' and 'id' args are required when creating a Chapter.
        The other fields are optional and have default values (see `help(tankobon.models.Chapter)`).

        Args:
            manga: The manga object.
        """

    @abc.abstractmethod
    def add_pages(self, chapter: models.Chapter):
        """Add pages to the chapter in the manga as a list of urls.
        The pages must be in ascending order.

        This method should assign pages to the chapter:

        ```python
        def pages(self, chapter):
            # do your parsing here
            chapter.pages = [...]  # assign directly to the chapter's pages.
        ```

        Args:
            chapter: The chapter object (already added to the manga).
        """

    def soup(self, url: str) -> bs4.BeautifulSoup:
        """Get a soup from a url.

        Args:
            url: The url to get a soup from.

        Returns:
            The soup of the url.
        """

        return utils.soup(url, session=self.session)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.domain = re.compile(cls.domain)
        cls.registered.append(cls)


class Cache(utils.PersistentDict):
    """A manga cache.

    Args:
        root: The root of the cache.

    Attributes:
        root: See args.
        alias: A map of manga url to manga hash.
    """

    INDEX = "index.json.gz"

    def __init__(self, root: Union[str, pathlib.Path] = utils.ROOT):
        if isinstance(root, str):
            root = pathlib.Path(root)

        self.root = root

        index = self.root / self.INDEX
        old_index = self.root / "index.json"

        if old_index.is_file():
            # indexes are now compressed by default so compress the old one
            with gzip.open(index, "wt") as f:
                f.write(old_index.read_text())

            old_index.unlink()

        super().__init__(self.root / self.INDEX, compress=True)

        self.alias: Dict[str, str] = {}

        # alias urls to their hashes
        for hash, manga in self.data.items():
            self.alias[manga["meta"].url] = hash

    def fullhash(self, part: str) -> str:
        """Get the full SHA512 hash of a manga when only given at least the first 8 letters of the hash.

        Args:
            part: The first 8 letters of the hash.

        Returns:
            The full hash, or an empty string if part was not found.

        Raises:
            ValueError, if the length part is less than 8.
        """
        if len(part) < SHORT_HASH_LEN:
            raise ValueError(f"part {part} is too short")

        for hash in self.data:
            if hash.startswith(part):
                return hash

        return ""

    def dump(self, manga: models.Manga):
        """Save this manga within the cache.

        Args:
            manga: The manga object to save.
        """

        self.data[manga.meta.hash] = manga.dump()

        self.alias[manga.meta.url] = manga.meta.hash

        (self.root / manga.meta.hash).mkdir(exist_ok=True)

    def load(self, hash: str) -> models.Manga:
        """Load a manga by its hash.

        Args:
            hash: The manga hash.

        Returns:
            The Manga object.

        Raises:
            MangaNotFoundError, if the manga does not exist in the cache.
        """

        if hash not in self.data:
            raise MangaNotFoundError(f"{hash} does not exist in cache")

        return models.Manga.load(self.data[hash])

    def delete(self, hash: str):
        """Delete a manga from the cache.

        Args:
            hash: The manga hash.

        Raises:
            MangaNotFoundError, if the manga does not exist in the cache.
        """
        if hash not in self.data:
            raise MangaNotFoundError(f"{hash} does not exist in cache")

        del self.alias[self.data[hash]["meta"].url]
        shutil.rmtree(str(self.root / hash))
        del self.data[hash]


class Downloader:
    """A manga downloader.

    Args:
        path: The path to where the manga chapters will be downloaded.
            For every manga chapter, a corrosponding folder is created if it does not exist.
    """

    MANIFEST = "manifest.json"

    def __init__(self, path: Union[str, pathlib.Path]):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self.path = path

        self.session = utils.UserSession()
        self.manifest = utils.PersistentDict(self.path / self.MANIFEST)

    def close(self):
        self.session.close()
        self.manifest.close()

    def download(
        self,
        chapter: models.Chapter,
        *,
        force: bool = False,
        progress: Optional[Callable[[int], None]] = None,
    ):
        """Download pages for a chapter.

        Args:
            chapter: The Chapter object to download.
            force: Whether or not to re-download the chapter if it already exists.
                Defaults to False.
            progress: A callback function which is called with the page number every time a page is downloaded.
                Defaults to None.

        Raises:
            PagesNotFoundError, if the chapter to be downloaded has no pages.
        """

        self.session.headers.update({"Referer": chapter.url})

        if not chapter.pages:
            raise PagesNotFoundError(f"chapter {chapter.id} does not have any pages")

        entry = self.manifest.setdefault(chapter.id, {})

        if chapter.lang in entry and not force:
            # bail out: dont re-download chapter
            return

        chapter.pages = cast(list, chapter.pages)

        chapter_path = self.path / chapter.id / chapter.lang
        chapter_path.mkdir(parents=True)

        pages = []
        total = len(chapter.pages)

        with cfutures.ThreadPoolExecutor(max_workers=8) as pool:
            futures = {
                pool.submit(self.session.get, url): count
                for count, url in enumerate(chapter.pages)
            }

            # progress count is different from page number as the page futures may not be in order.
            for p_count, future in enumerate(cfutures.as_completed(futures)):
                count = futures[future]

                _log.info(
                    f"downloader: [{chapter.id}] downloading page {count} of {total}"
                )

                try:
                    resp = future.result()
                    resp.raise_for_status()

                except requests.exceptions.RequestException as e:
                    _log.critical(f"downloader: failed to download page {count}: {e}")

                    # cant partially download, so remove chapter folder
                    shutil.rmtree(chapter_path)

                    raise e

                path = utils.save_response(chapter_path / str(count), resp)
                pages.append(str(path))

                if progress is not None:
                    progress(p_count)

        self.manifest[chapter.id][chapter.lang] = pages

    def download_cover(self, manga: models.Manga):
        """Download a manga's cover to the download path as 'cover.(ext)'.

        Args:
            manga: The manga to download a cover for.
        """

        self.session.headers.update({"Referer": manga.meta.url})

        with self.session.get(manga.meta.cover) as resp:
            utils.save_response(self.path / "cover", resp)

    def pdfify(
        self,
        chapters: List[str],
        dest: Union[str, pathlib.Path],
        lang: str = "en",
    ):
        """Create a PDF out of several (downloaded) chapters.
        The PDF will be A4 sized (vertical).

        Args:
            chapters: The chapters to create a PDF for.
            lang: The language of the chapters.
                Defaults to 'en'.
            dest: Where to write the PDF to.
        """

        document = fpdf.FPDF()

        for cid in natsort.natsorted(chapters):
            _log.info(f"pdf: adding chapter {cid}")

            pages = self.manifest[cid][lang]
            total = len(pages) - 1

            for page in natsort.natsorted(pages):
                _log.debug(f"adding page {page} of {total}")
                page_path = self.path / cid / lang / page

                width, height = imagesize.get(page_path)
                ratio = min(A4_WIDTH / width, A4_HEIGHT / height)

                document.add_page()
                document.image(str(page_path), 0, 0, w=width * ratio, h=height * ratio)

        document.output(str(dest), "F")

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()



# coding: utf8

import MangaDexPy  # type: ignore

from .. import core, models


class Parser(core.Parser):

    # mangadex has no website frontend yet, match base url plus manga id
    domain = r"mangadex\.org/([a-fA-F0-9\-]+)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client = MangaDexPy.MangaDex()
        self._cache = {}

    def _get_manga(self, url):
        manga_id = self.domain.search(url).group(1)

        if manga_id not in self._cache:
            self._cache[manga_id] = self.client.get_manga(manga_id)

        return self._cache[manga_id]

    def metadata(self, url):
        manga = self._get_manga(url)

        alt_titles = []

        for alt_title in manga.titles:
            alt_titles.extend(alt_title.values())

        authors = []

        # NOTE: array keys must have brackets '[]'!
        for author in self.client.search("author", params={"ids[]": manga.author}):
            authors.append(author.name)

        return models.Metadata(
            url=url,
            title=manga.title["en"],
            alt_titles=alt_titles,
            authors=authors,
            genres=[t.name["en"] for t in manga.tags],
            desc=manga.desc["en"],
            cover=self.client.get_cover(manga.cover).url,
        )

    def add_chapters(self, manga):
        manga_resp = self._get_manga(manga.meta.url)

        for chapter in manga_resp.get_chapters():

            # FIXME: il8n?
            if chapter.language == "en":

                manga.add(
                    models.Chapter(
                        id=chapter.chapter or "0",  # the chapter number
                        url=chapter.id,  # the mangadex chapter UUID
                        title=chapter.title,
                        volume=chapter.volume,
                    )
                )

    def add_pages(self, chapter):
        net_chapter = self.client.get_chapter(chapter.url).get_md_network()

        chapter.pages = net_chapter.pages
